
/*
 * $Id: ExpPortMgr.cc,v 1.12.2.1 1999/11/29 01:20:15 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include "Socket.h"
#include "Rnd.h"
#include "LogComment.h"
#include "SharedOpts.h"
#include "ErrorMgr.h"
#include "ExpPortMgr.h"

class PortHistory {
	friend ExpPortMgr;

	public:
		PortHistory();

		void configure(int aPortMin, int aPortMax);

		bool configured() const;
		bool belongs(int port) const;

		void record(int port, bool good);

	protected:
		bool usedPort(int port) const;

	protected:
		Ring<int> theValidPorts; // never- or successfully-used ports
		Ring<int> theVoidPorts;  // last-use-failed ports

		int thePortMin;
		int thePortMax;
};

static PortHistory TheHistory;



/* PortHistory */

PortHistory::PortHistory(): thePortMin(-1), thePortMax(-1) {
}

void PortHistory::configure(int aPortMin, int aPortMax) {
	Assert(!theValidPorts.count());

	thePortMin = aPortMin;
	thePortMax = aPortMax;

	const int count = thePortMax - thePortMin + 1;
	Assert(count >= 0);
	theValidPorts.resize(count);
	theVoidPorts.resize(count);

	Comment(6) << "port mgr is scanning ports " << thePortMin << ':' << thePortMax << " to find used ones" << endc;

	// push free ports, weed out used ones
	for (int p = thePortMin; p <= thePortMax; ++p)
		record(p, !usedPort(p));

	Comment(5) << "port mgr scanned " << count << " ports"
		<< " and found " << theVoidPorts.count() << " used ones" << endc;

	Assert(theValidPorts.count() + theVoidPorts.count() == count);

	// randomize the order in case there is a bad subrange or something...
	theValidPorts.randomize();
	theVoidPorts.randomize();
}

bool PortHistory::configured() const {
	return theValidPorts.count() + theVoidPorts.count() > 0;
}

void PortHistory::record(int port, bool good) {
	Assert(port >= thePortMin && port <= thePortMax);
	if (good)
		theValidPorts.enqueue(port);
	else
		theVoidPorts.enqueue(port);
}

bool PortHistory::belongs(int port) const {
	return port >= thePortMin && port <= thePortMax;
}

bool PortHistory::usedPort(int port) const {
	Socket s;
	if (s.create()) {
		bool used = !s.bind(NetAddr(0, port));
		s.close();
		return used;
	}
	return false;
}


/* ExpPortMgr */

ExpPortMgr::ExpPortMgr(const NetAddr &anAddr, int aPortMin, int aPortMax):
	PortMgr(anAddr) {
	theAddr.port(-1);

	if (!TheHistory.configured())
		TheHistory.configure(aPortMin, aPortMax);
}

int ExpPortMgr::allocPort(Socket &s) {
	int port;

	port = findPort(TheHistory.theValidPorts, s);
	if (port < 0)
		port = findPort(TheHistory.theVoidPorts, s);

	if (port < 0) {
		Comment << here << "no ports left at " << theAddr << ";"
			<< " bound: " << BoundLvl().level() 
			<< " valid: " << TheHistory.theValidPorts.count() 
			<< " void:  " << TheHistory.theVoidPorts.count() 
			<< " range: [" << TheHistory.thePortMin << ',' << TheHistory.thePortMax << ']'
			<< endc;
		Assert(TheHistory.theValidPorts.empty());
		errno = EADDRNOTAVAIL;
	} else {
		Assert(TheHistory.belongs(port));
	}
	return port;
}

void ExpPortMgr::freePort(int port, bool good) {
	TheHistory.record(port, good);
}

int ExpPortMgr::findPort(Ring<int> &ports, Socket &s) {
	// note: if ports == theVoidPorts, ports may "grow" inside the loop
	for (int i = ports.count(); i > 0; --i) {
		const int port = ports.dequeue();
		Assert(TheHistory.belongs(port));
		if (!bindToPort(s, port)) {
			if (ReportError(errno))
				Comment << "failed to bind to " << theAddr << ':' << port << endc;
			TheHistory.record(port, false);
		} else
			return port;
	}
	return -1;
}

bool ExpPortMgr::bindToPort(Socket &s, int port) const {
	NetAddr a(theAddr.addrN(), port);
	return s.bind(a);
}
