
/*
 * $Id: IOBuf.cc,v 1.10.2.2 1999/11/28 22:29:19 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include "Farm.h"
#include "LogComment.h"
#include "Rnd.h"
#include "IOBuf.h"
#include "gadgets.h"

class BufPool: public Farm<char> {
	public:
		BufPool(Size aBufCap = -1);
		virtual ~BufPool() { while (!empty()) destroy(get()); } // g++ bug: {}

		void bufCap(Size aBufCap);
		Size bufCap() const { return theBufCap; }

	protected:
		virtual char *gen();
		virtual void destroy(char *b) { delete[] b; }
		virtual void overflow(char *b) { push(b); }

	protected:
		Size theBufCap;
		Size theRepLvl; // when to report growth
};

static BufPool *TheBufPool = 0;

static char *TheRndBuf;           // pre-filled to rnd-ize other buffers
static int BufCapacity = 16*1024;  // XXX: hardcoded IO buf capacity


/* IOBuf */

IOBuf::IOBuf(): theBuf(0), theCapacity(0), theOutOff(0), theInOff(0) {
	Assert(TheBufPool);
	theCapacity = TheBufPool->bufCap();
	Assert(theCapacity > 0);
	unzip();
}

void IOBuf::appendRnd(Size sz) {
	Assert(spaceSize() >= sz);

	// randomize contents a bit
	static unsigned int rndOffset = 0;
	memcpy(space(), &TheRndBuf[rndOffset % (unsigned)BufCapacity], sz);
	rndOffset += 3989; // just some prime number

	appended(sz);
}

Size IOBuf::appendRndUpTo(Size size) {
	Assert(size >= 0);
	const Size sz = Min(spaceSize(), size);
	appendRnd(sz); // stuff with random data
	return sz;
}

void IOBuf::pack() {
	if (empty()) {
		reset();
	} else
	if (theOutOff) {
		memcpy(theBuf, content(), contSize());
		theInOff -= theOutOff;
		theOutOff = 0;
	}
}

// return [empty] buffer to shared area
void IOBuf::zip() {
	if (theBuf) { // could be zipped already
		TheBufPool->put(theBuf);
		theBuf = 0;
	}
}

void IOBuf::unzip() {
	if (!theBuf) {
		theBuf = TheBufPool->getDirty();
		Assert(theBuf);
	}
}


/* WrBuf */

void WrBuf::append(const char *buf, Size sz) {
	Assert(spaceSize() >= sz);
	memcpy(space(), buf, sz);
	appended(sz);
}


/* BufPool */

BufPool::BufPool(Size aBufCap):
	theBufCap(aBufCap), theRepLvl(Size::MB(20)) {
	Assert(TheRndBuf);
}

void BufPool::bufCap(Size aBufCap) {
	Assert(empty() && !outLevel());
	theBufCap = aBufCap;
}

char *BufPool::gen() {
	Assert(theBufCap > 0);
	char *b = new char[theBufCap];

	const Size poolSize = outLevel()*theBufCap;
	if (poolSize >= theRepLvl) {
		Comment(1) << "buffer pool grew to " 
			<< outLevel() << " x " << theBufCap << " = " << poolSize << endc;
		theRepLvl += Min(theRepLvl, Size::MB(10));
	}

	return b;
}


/* initialization */

int IOBufInit::TheUseCount = 0;

void IOBufInit::init() {
	RndGen rng;
	const int cap = BufCapacity * 2 + 1; // more to allow random offsets
	TheRndBuf = new char[cap];
	for (int i = 0; i < cap; ++i) {
		const char c = (char)rng(33, 127);
		// to assist with HTML parsing, substitute '<' with a space
		TheRndBuf[i] = c == '<' ? ' ' : c; 
	}

	TheBufPool = new BufPool(BufCapacity);
}

void IOBufInit::clean() {
	delete TheBufPool;
	TheBufPool = 0;
	delete[] TheRndBuf;
}

