
/*
 * $Id: PhaseLx.cc,v 1.11.4.3 1999/11/20 18:10:54 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <iomanip.h>

#include "PhaseLx.h"


PhaseLx::PhaseLx(): theExor(0) {
}

bool PhaseLx::configure(const String &aName, const LxParams &params) {
	if (IntvlLx::configure(aName, params))
		return true;

	/* check parametrized objects first */

	if (theName == "rptm_hist" && params.count()) {
		checkParams(params, 1);
		if (params[0] == "hit")
			theExor = &PhaseLx::exHitRptmH;
		else
		if (params[0] == "miss")
			theExor = &PhaseLx::exMissRptmH;
		else
		if (params[0] == "all")
			theExor = &PhaseLx::exAllRptmH;
		else
			badParam(params, 0, "hit, miss, or all");

		return theExor;
	}

	/* the rest of objects are parameterless */

	if (theName == "phase_name")
		theExor = &PhaseLx::exName;
	else
	if (theName == "conn_use_hist")
		theExor = &PhaseLx::exConnUseH;
	else
	if (theName == "conn_ttl_hist")
		theExor = &PhaseLx::exConnTtlH;
	else
	if (theName == "conn_close_act_use_hist")
		theExor = &PhaseLx::exConnCloseActUH;
	else
	if (theName == "conn_close_act_ttl_hist")
		theExor = &PhaseLx::exConnCloseActTH;
	else
	if (theName == "conn_close_frgn_use_hist")
		theExor = &PhaseLx::exConnCloseFrgnUH;
	else
	if (theName == "conn_close_frgn_ttl_hist")
		theExor = &PhaseLx::exConnCloseFrgnTH;
	else
	if (theName == "conn_close_lcl_use_hist")
		theExor = &PhaseLx::exConnCloseLclUH;
	else
	if (theName == "conn_close_lcl_ttl_hist")
		theExor = &PhaseLx::exConnCloseLclTH;
	else
	if (theName == "soread_sz_hist")
		theExor = &PhaseLx::exSoReadSzH;
	else
	if (theName == "sowrite_sz_hist")
		theExor = &PhaseLx::exSoWriteSzH;
	else
	if (theName == "rep_sz_hist")
		theExor = &PhaseLx::exRepSzH;
	else
	if (theName == "rptm_hist")
		theExor = &PhaseLx::exAllRptmH; // parameterless case
	else
	if (theName == "rptm_vs_sz")
		theExor = &PhaseLx::exRptmVsSz;
	else
	if (theName == "conn_ttl_vs_use")
		theExor = &PhaseLx::exConnTtlVsUse;
	else
		return false;

	checkParams(params, 0);

	return theExor;
}

void PhaseLx::extract(const LxRec &r, ostream &os) {
	if (theExor)
		(this->*theExor)(r.phase(), os);
	else
		IntvlLx::extract(r, os);
}

void PhaseLx::printHistogram(const Histogram &h, ostream &os) const {
	h.report(0.01, os);
	os << endl;
	h.stats().print(os, theName + ":");
}

void PhaseLx::printVsLine(const Histogram &h, int pStep, ostream &os) const {
	// print count and mean
	os << ' ' << setw(6) << h.stats().count()
		<< ' ' << setw(5) << h.stats().mean();

	// get and print percentiles
	Array<HistogramBin> percs;
	Percentiles(h, percs);
	for (int i = pStep; i <= 100; i += pStep) {
		if (i <= percs.count()) // do we ever have less?
			os << ' ' << setw(3) << percs[i-1].sup-1;
		else
			os << ' ' << setw(3) << '?';
	}
}

void PhaseLx::exConnUseH(const StatPhaseRec &p, ostream &os) const {
	Log2Hist h;
	p.theConnClose.useHist(h);
	printHistogram(h, os);
}

void PhaseLx::exConnTtlH(const StatPhaseRec &p, ostream &os) const {
	TimeHist h;
	p.theConnClose.ttlHist(h);
	printHistogram(h, os);
}

void PhaseLx::exAllRptmH(const StatPhaseRec &p, ostream &os) const {
	TimeHist h;
	h.add(p.theHitRptmH);
	h.add(p.theMissRptmH);
	printHistogram(h, os);
}

void PhaseLx::printTimeVsHist(const TimeVsSmthHist &vsh, int min, const String &type, ostream &os) const {
	const int pStep = 5; // 5%;

	// header
	os << endl 
		<< "#   " << type << "     response_time" << endl
		<< "# min   max  count  mean";
	for (int i = pStep; i <= 100; i += pStep)
		os << ' ' << setw(2) << i << '%';
	os << endl;

	// for each size group, print mean and percentiles
	const int lastGrp = vsh.lastGrpId();
	for (int g = 0; g <= lastGrp; ++g) {
		int max = vsh.groupMax(g);
		const Histogram &h = vsh.groupHist(g);

		if (!h.stats().count())
			continue;

		os << setw(5) << min << ' ' << setw(5) << max;
		printVsLine(h, pStep, os);
		os << endl;

		min = max + 1;
	}
}

void PhaseLx::exRptmVsSz(const StatPhaseRec &p, ostream &os) const {
	printTimeVsHist(p.theRptmVsSz, 0, "size", os);
}

void PhaseLx::exConnTtlVsUse(const StatPhaseRec &p, ostream &os) const {
	printTimeVsHist(p.theConnLifeVsUse, 1, "uses", os);
}
