
/*
 * $Id: Ring.h,v 1.4.4.3 1999/11/29 03:12:40 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#ifndef X_RING_H
#define X_RING_H

#include "Array.h"
#include "Rnd.h"

// FIFO Queue of <Item>s with fixed capacity (i.e., ring)
// note dynamic resize()-ing on enqueue() does work but is disabled

template <class Item>
class Ring: protected Array<Item> {
	public:
		Ring(int aCapacity = 0): Array<Item>(aCapacity), theInOff(0), theOutOff(0) {}

		int capacity() const { return theCapacity; }
		int count() const { return theInOff - theOutOff; }
		bool empty() const { return theInOff <= theOutOff; }
		bool full() const { return count() >= theCapacity; }

		void enqueue(Item i) { Assert(!full()); item(theInOff++ % theCapacity) = i; check(); }
		Item dequeue() { return item(theOutOff++ % theCapacity); }

		void reset() { theInOff = theOutOff = 0; }
		inline void resize(int aCap);

		inline void randomize();

	protected:
		inline void check();

	protected:
		int theInOff;
		int theOutOff;
};


/* implementaion of in-lined methods */

template <class Item>
inline
void Ring<Item>::randomize() {
	RndGen rng;
	for (int i = theOutOff; i < theInOff; ++i)
		Array<Item>::swap(i % theCapacity,
			rng(theOutOff, theInOff) % theCapacity);
}

template <class Item>
inline
void Ring<Item>::resize(int aCap) {
	const int oldCap = theCapacity;
	Item *oldItems = theItems;
	theItems = 0;

	grow(aCap);

	if (oldItems) {
		if (count()) {
			const int old_hi = theInOff > oldCap ? oldCap : theInOff;
			const int gap = theCapacity - old_hi;
			const int new_low = theOutOff + gap;
			if (old_hi > theOutOff)
				memmove(new_low, oldItems+theOutOff, old_hi-theOutOff);
			if (theInOff > oldCap)
				memmove(0, oldItems, theInOff-oldCap);
			theInOff += gap;
			theOutOff += gap;
		}
		delete[] oldItems;
	}
}

template <class Item>
inline
void Ring<Item>::check() {
	Assert(theInOff > theOutOff && count() <= theCapacity);
	if (theOutOff >= theCapacity) {
		theInOff %= theCapacity;
		theOutOff %= theCapacity;
		if (theInOff <= theOutOff)
			theInOff += theCapacity;
	}
}

#endif
