
/*
 * $Id: msl_test.c,v 1.2.2.3 1999/10/28 06:00:39 rousskov Exp $
 *
 * Msl_test attempts to estimate the maximum time a remote server 
 * keeps a TCP connection in a TIME_WAIT state. To run the test,
 * one must have root access. The tool requires BPF package and
 * runs only on FreeBSD.
 *
 * Author: Glenn Chisholm <glenn@ircache.net>
 *
 */

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/uio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/bpf.h>
#include <net/route.h>
#include <net/ethernet.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

#include <arpa/inet.h>

static int validTCPPacket(struct tcphdr *tcphdr);
static int createTCPsession(unsigned long server, int port);
static int setBPFServerFilter(int bpf, \
		unsigned long server, \
    		unsigned long localhost, \
		int port);
static int openBPFfilter(char *interface, \
		u_int * buf_len, \
    		unsigned long server, \
		unsigned long localhost, \
		int port);
static void determineMSLWithPacket(int bpf_sock, \
    		struct ether_header *ether_header, \
		u_short bh_caplen, \
		u_int buf_len);
static void breakupPacket(void *buf, \
        	struct bpf_hdr **bpf_hdr, \
        	struct ether_header **ether_header, \
        	struct ip **ip, \
        	struct tcphdr **tcphdr);
static void printPacket(struct ip *ip, struct tcphdr *tcphdr);
static void parseOptions(int argc, char *argv[]);
static int deleteRoute(int route, unsigned long server);
static int addRoute(unsigned long server);
static int getPacketType(void *buf);
static int chooseBPFint(void);
static void usage(const char *prg_name);

#define READ_SIZE 1024
#define SEQ_ADD_ROUTE 100
#define SEQ_DELETE_ROUTE 200
#define HTTP_REQUEST "GET / HTTP/1.0\n\n"
#define SIZEOF_BPF_HDR  (sizeof(struct bpf_hdr) <= 20 ? 18 : \
    sizeof(struct bpf_hdr))

static char *arg_interface = NULL, 
	*arg_dst_host = NULL, 
	*arg_source_host = NULL;
static int arg_port = 0;

static void
usage(const char *prg_name)
{
    fprintf(stderr,
	"Usage: %s -i interface-name -s src-host -d dst-host -p dst-port\n"
	"       -i interface-name The name of the interface which to listen.\n"
	"       -s src-host	 The IP address of the local host\n"
	"       -d dst-host       The IP address of the destination.\n"
	"       -p dst-port       Port which to connect to.\n"
	"You must be root to run this program.\n",
	prg_name);
    exit(1);
}

static void
parseOptions(int argc, char *argv[])
{
    extern char *optarg;
    int c;

    while ((c = getopt(argc, argv, "i:s:d:p:?")) != -1) {
	switch (c) {
	case 'i':
	    arg_interface = strdup(optarg);
	    break;
	case 's':
	    arg_source_host = strdup(optarg);
	    break;
	case 'd':
	    arg_dst_host = strdup(optarg);
	    break;
	case 'p':
	    arg_port = atoi(optarg);
	    break;
	case '?':
	default:
	    usage(argv[0]);
	    break;
	}
    }
}


int
main(int argc, char **argv)
{
    void *buf = NULL, *p = NULL;
    u_int buf_len = 0;
    struct ip *ip = NULL;
    struct tcphdr *tcphdr = NULL;
    struct bpf_hdr *bpf_hdr = NULL;
    struct ether_header *ether_header = NULL;
    unsigned long server, localhost;
    int bpf_sock, invalid, route;

    if (argc == 9) {
	parseOptions(argc, argv);
    } else {
	usage(argv[0]);
    }

    if (getuid()) {
	printf("You must be root to run this program!\n");
	return(-1);
    }

    printf("Determining MSL of %s ...\n", arg_dst_host);
    server = inet_addr(arg_dst_host);
    localhost = inet_addr(arg_source_host);

    if ((bpf_sock = openBPFfilter(arg_interface,
		&buf_len,
		server,
		localhost,
		arg_port)) < 0)
	return(errno);

    if (createTCPsession(server, arg_port))
	return errno;

    buf = (void *) malloc(sizeof(void) * buf_len);
    memset(buf, '\0', buf_len);
    if (read(bpf_sock, buf, buf_len) < 0)
	return(errno);

    p = buf;
    invalid = 1;
    while (invalid) {
        breakupPacket(p, &bpf_hdr, &ether_header, &ip, &tcphdr);
	if (validTCPPacket(tcphdr)) {
	    invalid = 0;
	} else {
	    p = (void *) p + BPF_WORDALIGN(bpf_hdr->bh_hdrlen + bpf_hdr->bh_caplen);
	}
    }

    if((route = addRoute(server)) < 0) {
	printf("Unable to create blackhole route\n");
	return(errno);
    }

    printf("Sending packets -> ");
    printPacket(ip, tcphdr);
    if (setBPFServerFilter(bpf_sock, server, localhost, arg_port))
	return(errno);

    determineMSLWithPacket(bpf_sock,
	ether_header,
	bpf_hdr->bh_caplen,
	buf_len);

    if(deleteRoute(route, server) < 0) 
        printf("Unable to remove blackhole route\n");

    free(buf);
    close(bpf_sock);
}

void
breakupPacket(void *buf, 
	struct bpf_hdr **bpf_hdr,
	struct ether_header **ether_header,
	struct ip **ip,
	struct tcphdr **tcphdr)
{
    int ip_hl;

    *bpf_hdr = (struct bpf_hdr *) buf;
    *ether_header = (struct ether_header *) ((void *) *bpf_hdr + (*bpf_hdr)->bh_hdrlen);
    *ip = (struct ip *) ((void *) *ether_header + sizeof(**ether_header));
    ip_hl = (*ip)->ip_hl << 2;
    *tcphdr = (struct tcphdr *) ((void *) *ip + ip_hl);
}

int
getPacketType(void *buf)
{
    struct ip *ip = NULL;
    struct tcphdr *tcphdr = NULL;
    struct bpf_hdr *bpf_hdr = NULL;
    struct ether_header *ether_header = NULL;

    breakupPacket(buf, &bpf_hdr, &ether_header, &ip, &tcphdr);

    if (tcphdr->th_flags & TH_SYN)
        return (1); 

    return(0);
}

int
addRoute(unsigned long server)
{
    struct {
	struct rt_msghdr rtm;
	struct sockaddr_in host, localhost;
    } rtmsg;
    int route, seq = SEQ_ADD_ROUTE, msglen;

    if ((route = socket(PF_ROUTE, SOCK_RAW, 0)) < 0)
	return (-1);

    msglen = sizeof(rtmsg);
    memset(&rtmsg, 0, msglen);

    rtmsg.host.sin_family = AF_INET;
    rtmsg.host.sin_addr.s_addr = server;
    rtmsg.host.sin_len = 16;

    rtmsg.localhost.sin_family = AF_INET;
    rtmsg.localhost.sin_addr.s_addr = inet_addr("127.0.0.1");
    rtmsg.localhost.sin_len = 16;

    rtmsg.rtm.rtm_msglen = msglen;
    rtmsg.rtm.rtm_version = RTM_VERSION;
    rtmsg.rtm.rtm_type = RTM_ADD;
    rtmsg.rtm.rtm_flags = RTF_UP | RTF_STATIC | RTA_GATEWAY | RTF_HOST | RTF_BLACKHOLE;
    rtmsg.rtm.rtm_addrs = RTA_DST | RTA_GATEWAY;
    rtmsg.rtm.rtm_pid = getpid();
    rtmsg.rtm.rtm_seq = seq;

    if (write(route, (char *) &rtmsg, msglen) < 0) 
	return (-1);
    return (route);
}

int
deleteRoute(int route, unsigned long server)
{
    struct {
        struct rt_msghdr rtm;
        struct sockaddr_in host;
    } rtmsg;
    int seq = SEQ_DELETE_ROUTE, msglen;             

    msglen = sizeof(rtmsg);         
    memset(&rtmsg, 0, msglen); 
    
    rtmsg.host.sin_family = AF_INET;
    rtmsg.host.sin_addr.s_addr = server;
    rtmsg.host.sin_len = 16;
    
    rtmsg.rtm.rtm_msglen = msglen;  
    rtmsg.rtm.rtm_version = RTM_VERSION;
    rtmsg.rtm.rtm_type = RTM_DELETE;
    rtmsg.rtm.rtm_addrs = RTA_DST;
    rtmsg.rtm.rtm_pid = getpid();
    rtmsg.rtm.rtm_seq = seq;
    
    if (write(route, (char *) &rtmsg, msglen) < 0)        
        return (-1);
    close(route);
    return(1);
}


void
printPacket(struct ip *ip, struct tcphdr *tcphdr)
{
    printf("%s.%d > ", inet_ntoa(ip->ip_src), ntohs(tcphdr->th_sport));
    printf("%s.%d\t", inet_ntoa(ip->ip_dst), ntohs(tcphdr->th_dport));

    if (tcphdr->th_flags & TH_SYN)
	printf("SYN ");

    if (tcphdr->th_flags & TH_FIN)
	printf("FIN ");

    if (tcphdr->th_flags & TH_PUSH)
	printf("PUSH ");

    if (tcphdr->th_flags & TH_ACK)
	printf("ACK ");

    if (tcphdr->th_flags & TH_RST)
	printf("RST");

    printf("\n");
}

void
determineMSLWithPacket(int bpf_sock,
    struct ether_header *ether_header,
    u_short bh_caplen,
    u_int buf_len)
{
    void *buf = NULL;
    fd_set readfds;
    struct timeval timeout;
    int nfds = bpf_sock + 1, invalid = 0, seconds = 1;

    FD_ZERO(&readfds);
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;
    buf = (void *) malloc(sizeof(void) * buf_len);
    while (!invalid) {
	FD_SET(bpf_sock, &readfds);
	if (write(bpf_sock, ether_header, bh_caplen) < 0)
	    return;
	if (select(nfds, &readfds, NULL, NULL, &timeout)) {
	    if (read(bpf_sock, buf, buf_len) < 0)
		return;
	    if(getPacketType((void *)buf)) 
	        invalid = 1;
	    else {
		seconds++;
		sleep(1);
	    }
	} else {
	    seconds++;
	}
    }
    free(buf);
    printf("TCP TIME_WAIT of %d\n", seconds);
}

int
setBPFServerFilter(int bpf, unsigned long server, unsigned long localhost, int port)
{
    u_int immediate = 1;
    struct bpf_program bpf_program;
    struct bpf_insn insns[] =
    {
	BPF_STMT(BPF_LD + BPF_H + BPF_ABS, 12),
	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, ETHERTYPE_IP, 0, 12),
	BPF_STMT(BPF_LD + BPF_B + BPF_ABS, 23),
	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, IPPROTO_TCP, 0, 10),
	BPF_STMT(BPF_LD + BPF_W + BPF_ABS, 26),
	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, htonl(server), 0, 8),
	BPF_STMT(BPF_LD + BPF_W + BPF_ABS, 30),
	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, htonl(localhost), 0, 5),
	BPF_STMT(BPF_LD + BPF_H + BPF_ABS, 20),
	BPF_JUMP(BPF_JMP + BPF_JSET + BPF_K, 0x1fff, 4, 0),
	BPF_STMT(BPF_LDX + BPF_B + BPF_MSH, 14),
	BPF_STMT(BPF_LD + BPF_H + BPF_IND, 14),
	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, port, 0, 1),
	BPF_STMT(BPF_RET + BPF_K, (u_int) - 1),
	BPF_STMT(BPF_RET + BPF_K, 0),
    };

    bpf_program.bf_insns = insns;
    bpf_program.bf_len = 15;
    if (ioctl(bpf, BIOCSETF, &bpf_program))
	return (-1);

    if (ioctl(bpf, BIOCIMMEDIATE, &immediate))
	return (-1);

    if (ioctl(bpf, BIOCFLUSH, NULL))
	return (-1);

    return (0);
}

int
validTCPPacket(struct tcphdr *tcphdr)
{
    if (tcphdr->th_flags & TH_SYN)
	return (1);

    if (tcphdr->th_flags & TH_FIN)
	return (0);

    if (tcphdr->th_flags & TH_PUSH)
	return (0);

    if (tcphdr->th_flags & TH_ACK)
	return (0);

    if (tcphdr->th_flags & TH_RST)
	return (0);

    return (0);
}

int
openBPFfilter(char *interface, u_int * buf_len, unsigned long server, unsigned long localhost, int port)
{
    int bpf;
    struct ifreq *ifreq = NULL;
    struct timeval timeval;
    struct bpf_program bpf_program;
    struct bpf_insn insns[] =
    {
	BPF_STMT(BPF_LD + BPF_H + BPF_ABS, 12),
	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, ETHERTYPE_IP, 0, 12),
	BPF_STMT(BPF_LD + BPF_B + BPF_ABS, 23),
	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, IPPROTO_TCP, 0, 10),
	BPF_STMT(BPF_LD + BPF_W + BPF_ABS, 26),
	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, htonl(localhost), 0, 8),
	BPF_STMT(BPF_LD + BPF_W + BPF_ABS, 30),
	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, htonl(server), 0, 5),
	BPF_STMT(BPF_LD + BPF_H + BPF_ABS, 20),
	BPF_JUMP(BPF_JMP + BPF_JSET + BPF_K, 0x1fff, 4, 0),
	BPF_STMT(BPF_LDX + BPF_B + BPF_MSH, 14),
	BPF_STMT(BPF_LD + BPF_H + BPF_IND, 16),
	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, port, 0, 1),
	BPF_STMT(BPF_RET + BPF_K, (u_int) - 1),
	BPF_STMT(BPF_RET + BPF_K, 0),
    };

    bpf = chooseBPFint();

    ifreq = (struct ifreq *) malloc(sizeof(struct ifreq));
    strcpy(ifreq->ifr_name, interface);
    if (ioctl(bpf, BIOCSETIF, ifreq))
	return (-1);
    free(ifreq);

    if (ioctl(bpf, BIOCGBLEN, buf_len))
	return (-1);

    timeval.tv_sec = 1;
    timeval.tv_usec = 0;
    if (ioctl(bpf, BIOCSRTIMEOUT, &timeval))
	return (-1);

    bpf_program.bf_insns = insns;
    bpf_program.bf_len = 15;
    if (ioctl(bpf, BIOCSETF, &bpf_program))
	return (-1);

    if (ioctl(bpf, BIOCFLUSH, NULL))
	return (-1);

    return (bpf);
}

int
chooseBPFint(void)
{
    int bpf, i;
    char p[32];

    for (i = 0; i < 4; i++) {
        snprintf(p, 32, "/dev/bpf%d", i);
        bpf = open(p, O_RDWR);
        if (bpf >= 0)
            return (bpf);
        perror(p);  
    }   
    exit(1);

}

int
createTCPsession(unsigned long server, int port)
{
    int sock;
    struct sockaddr_in servaddr;
    char buf[READ_SIZE];

    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	return (1);

    servaddr.sin_family = AF_INET;
    servaddr.sin_port = htons(port);
    servaddr.sin_addr.s_addr = server;

    if (connect(sock, (struct sockaddr *) &servaddr, sizeof(servaddr))) {
	perror("connect");
	return (1);
    }

    write(sock, HTTP_REQUEST, strlen(HTTP_REQUEST));

    while (read(sock, buf, READ_SIZE)) {

    }
    sleep(1);
    close(sock);
    return (0);
}
